@@ -1,30 +1,34 @@
 /*
  * pg_bulkload: lib/pg_btree.c
  *
- *	  Copyright(C) 2007-2009, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ *	  Copyright (c) 2007-2010, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 /**
  * @file
  * @brief implementation of B-Tree index processing module
  */
-#include "postgres.h"
+#include "pg_bulkload.h"
 
 #include "access/genam.h"
 #include "access/heapam.h"
+#include "access/nbtree.h"
 #include "access/xact.h"
 #include "catalog/index.h"
 #include "executor/executor.h"
+#include "storage/fd.h"
 #include "storage/lmgr.h"
+#include "storage/smgr.h"
+#include "utils/builtins.h"
 #include "utils/lsyscache.h"
+#include "utils/memutils.h"
 #include "utils/tqual.h"
 
 #if PG_VERSION_NUM >= 80400
 #include "utils/snapmgr.h"
 #endif
 
-#include "pg_bulkload_win32.h"
-#include "pg_btree.h"
+#include "logger.h"
 
 static BTSpool *unused_bt_spoolinit(Relation, bool, bool);
 static void unused_bt_spooldestroy(BTSpool *);
@@ -36,15 +40,28 @@ static void unused_bt_leafbuild(BTSpool *, BTSpool *);
 #define _bt_spool			unused_bt_spool
 #define _bt_leafbuild		unused_bt_leafbuild
 
-#include "nbtsort.c"
+#if PG_VERSION_NUM >= 90200
+#error unsupported PostgreSQL version
+#elif PG_VERSION_NUM >= 90100
+#include "nbtree/nbtsort-9.1.c"
+#elif PG_VERSION_NUM >= 90000
+#include "nbtree/nbtsort-9.0.c"
+#elif PG_VERSION_NUM >= 80400
+#include "nbtree/nbtsort-8.4.c"
+#elif PG_VERSION_NUM >= 80300
+#include "nbtree/nbtsort-8.3.c"
+#else
+#error unsupported PostgreSQL version
+#endif
 
 #undef _bt_spoolinit
 #undef _bt_spooldestroy
 #undef _bt_spool
 #undef _bt_leafbuild
 
-#include "pg_bulkload.h"
+#include "pg_btree.h"
 #include "pg_profile.h"
+#include "pgut/pgut-be.h"
 
 /**
  * @brief Reader for existing B-Tree index
@@ -60,22 +77,97 @@ typedef struct BTReader
 	char			   *page;	/**< Cached page */
 } BTReader;
 
+static BTSpool **IndexSpoolBegin(ResultRelInfo *relinfo, bool enforceUnique);
+static void IndexSpoolEnd(Spooler *self , bool reindex);
+static void IndexSpoolInsert(BTSpool **spools, TupleTableSlot *slot, ItemPointer tupleid, EState *estate, bool reindex);
+
+static IndexTuple BTSpoolGetNextItem(BTSpool *spool, IndexTuple itup, bool *should_free);
 static bool BTReaderInit(BTReader *reader, Relation rel);
 static void BTReaderTerm(BTReader *reader);
 static void BTReaderReadPage(BTReader *reader, BlockNumber blkno);
 static IndexTuple BTReaderGetNextItem(BTReader *reader);
 
-static void _bt_mergebuild(BTSpool *btspool, Relation heapRel, bool use_wal);
-static void _bt_mergeload(BTWriteState *wstate, BTSpool *btspool,
+static void _bt_mergebuild(Spooler *self, BTSpool *btspool);
+static void _bt_mergeload(Spooler *self, BTWriteState *wstate, BTSpool *btspool,
 						  BTReader *btspool2, Relation heapRel);
+static int compare_indextuple(const IndexTuple itup1, const IndexTuple itup2,
+	ScanKey entry, int keysz, TupleDesc tupdes, bool *hasnull);
 static bool heap_is_visible(Relation heapRel, ItemPointer htid);
-static void report_unique_violation(Relation rel, IndexTuple itup);
+static void remove_duplicate(Spooler *self, Relation heap, IndexTuple itup, const char *relname);
+
+
+void
+SpoolerOpen(Spooler *self,
+			Relation rel,
+			bool use_wal,
+			const WriterOptions *options)
+{
+	memset(self, 0, sizeof(Spooler));
+
+	self->on_duplicate = options->on_duplicate;
+	self->use_wal = use_wal;
+	self->max_dup_errors = options->max_dup_errors;
+	self->dup_old = 0;
+	self->dup_new = 0;
+	self->dup_badfile = pstrdup(options->dup_badfile);
+	self->dup_fp = NULL;
+
+	self->relinfo = makeNode(ResultRelInfo);
+	self->relinfo->ri_RangeTableIndex = 1;	/* dummy */
+	self->relinfo->ri_RelationDesc = rel;
+	self->relinfo->ri_TrigDesc = NULL;	/* TRIGGER is not supported */
+	self->relinfo->ri_TrigInstrument = NULL;
+
+	ExecOpenIndices(self->relinfo);
+
+	self->estate = CreateExecutorState();
+	self->estate->es_num_result_relations = 1;
+	self->estate->es_result_relations = self->relinfo;
+	self->estate->es_result_relation_info = self->relinfo;
+
+	self->slot = MakeSingleTupleTableSlot(RelationGetDescr(rel));
+
+	self->spools = IndexSpoolBegin(self->relinfo,
+								   options->max_dup_errors == 0);
+}
+
+void
+SpoolerClose(Spooler *self)
+{
+	/* Merge indexes */
+	if (self->spools != NULL)
+		IndexSpoolEnd(self, true);
+
+	/* Terminate spooler. */
+	ExecDropSingleTupleTableSlot(self->slot);
+	if (self->estate->es_result_relation_info)
+		ExecCloseIndices(self->estate->es_result_relation_info);
+	FreeExecutorState(self->estate);
+
+	/* Close and release members. */
+	if (self->dup_fp != NULL && FreeFile(self->dup_fp) < 0)
+		ereport(WARNING,
+				(errcode_for_file_access(),
+				 errmsg("could not close duplicate bad file \"%s\": %m",
+						self->dup_badfile)));
+	if (self->dup_badfile != NULL)
+		pfree(self->dup_badfile);
+}
+
+void
+SpoolerInsert(Spooler *self, HeapTuple tuple)
+{
+	/* Spool keys in the tuple */
+	ExecStoreTuple(tuple, self->slot, InvalidBuffer, false);
+	IndexSpoolInsert(self->spools, self->slot, &(tuple->t_self), self->estate, true);
+	BULKLOAD_PROFILE(&prof_writer_index);
+}
 
 /*
  * IndexSpoolBegin - Initialize spools.
  */
-BTSpool **
-IndexSpoolBegin(ResultRelInfo *relinfo)
+static BTSpool **
+IndexSpoolBegin(ResultRelInfo *relinfo, bool enforceUnique)
 {
 	int				i;
 	int				numIndices = relinfo->ri_NumIndices;
@@ -91,7 +183,10 @@ IndexSpoolBegin(ResultRelInfo *relinfo)
 		{
 			elog(DEBUG1, "pg_bulkload: spool \"%s\"",
 				RelationGetRelationName(indices[i]));
-			spools[i] = _bt_spoolinit(indices[i], indices[i]->rd_index->indisunique, false);
+			spools[i] = _bt_spoolinit(indices[i],
+					enforceUnique ? indices[i]->rd_index->indisunique: false,
+					false);
+			spools[i]->isunique = indices[i]->rd_index->indisunique;
 		}
 		else
 			spools[i] = NULL;
@@ -104,26 +199,21 @@ IndexSpoolBegin(ResultRelInfo *relinfo)
  * IndexSpoolEnd - Flush and delete spools.
  */
 void
-IndexSpoolEnd(BTSpool **spools,
-			  ResultRelInfo *relinfo,
-			  bool reindex,
-			  bool use_wal)
+IndexSpoolEnd(Spooler *self, bool reindex)
 {
+	BTSpool **spools = self->spools;
 	int				i;
-	RelationPtr		indices = relinfo->ri_IndexRelationDescs;
+	RelationPtr		indices = self->relinfo->ri_IndexRelationDescs;
 
 	Assert(spools != NULL);
-	Assert(relinfo != NULL);
+	Assert(self->relinfo != NULL);
 
-	for (i = 0; i < relinfo->ri_NumIndices; i++)
+	for (i = 0; i < self->relinfo->ri_NumIndices; i++)
 	{
 		if (spools[i] != NULL)
 		{
-			BULKLOAD_PROFILE_PUSH();
-			_bt_mergebuild(spools[i], relinfo->ri_RelationDesc, use_wal);
-			BULKLOAD_PROFILE_POP();
+			_bt_mergebuild(self, spools[i]);
 			_bt_spooldestroy(spools[i]);
-			BULKLOAD_PROFILE(&prof_index_merge);
 		}
 		else if (reindex)
 		{
@@ -132,9 +222,9 @@ IndexSpoolEnd(BTSpool **spools,
 			/* Close index before reindex to pass CheckTableNotInUse. */
 			relation_close(indices[i], NoLock);
 			indices[i] = NULL;
-			reindex_index(indexOid);
+			reindex_index(indexOid, false);
 			CommandCounterIncrement();
-			BULKLOAD_PROFILE(&prof_index_reindex);
+			BULKLOAD_PROFILE(&prof_reindex);
 		}
 		else
 		{
@@ -150,7 +240,7 @@ IndexSpoolEnd(BTSpool **spools,
  *
  *	Copy from ExecInsertIndexTuples.
  */
-void
+static void
 IndexSpoolInsert(BTSpool **spools, TupleTableSlot *slot, ItemPointer tupleid, EState *estate, bool reindex)
 {
 	ResultRelInfo  *relinfo;
@@ -194,11 +284,9 @@ IndexSpoolInsert(BTSpool **spools, TupleTableSlot *slot, ItemPointer tupleid, ES
 
 		indexInfo = indexInfoArray[i];
 
-#if PG_VERSION_NUM >= 80300
 		/* If the index is marked as read-only, ignore it */
 		if (!indexInfo->ii_ReadyForInserts)
 			continue;
-#endif
 
 		/* Check for partial index */
 		if (indexInfo->ii_Predicate != NIL)
@@ -243,10 +331,12 @@ IndexSpoolInsert(BTSpool **spools, TupleTableSlot *slot, ItemPointer tupleid, ES
 
 
 static void
-_bt_mergebuild(BTSpool *btspool, Relation heapRel, bool use_wal)
+_bt_mergebuild(Spooler *self, BTSpool *btspool)
 {
+	Relation heapRel = self->relinfo->ri_RelationDesc;
 	BTWriteState	wstate;
 	BTReader		reader;
+	bool			merge;
 
 	Assert(btspool->index->rd_index->indisvalid);
 
@@ -258,7 +348,7 @@ _bt_mergebuild(BTSpool *btspool, Relation heapRel, bool use_wal)
 	 * We need to log index creation in WAL iff WAL archiving is enabled AND
 	 * it's not a temp index.
 	 */
-	wstate.btws_use_wal = use_wal &&
+	wstate.btws_use_wal = self->use_wal &&
 		XLogArchivingActive() && !wstate.index->rd_istemp;
 
 	/* reserve the metapage */
@@ -273,72 +363,68 @@ _bt_mergebuild(BTSpool *btspool, Relation heapRel, bool use_wal)
 	 */
 	LockRelation(wstate.index, AccessExclusiveLock);
 	FlushRelationBuffers(wstate.index);
-	BULKLOAD_PROFILE(&prof_index_merge_flush);
+	BULKLOAD_PROFILE(&prof_flush);
 
-	PG_TRY();
+	merge = BTReaderInit(&reader, wstate.index);
+
+	elog(DEBUG1, "pg_bulkload: build \"%s\" %s merge (%s wal)",
+		RelationGetRelationName(wstate.index),
+		merge ? "with" : "without",
+		wstate.btws_use_wal ? "with" : "without");
+
+	/* Assign a new file node. */
+	RelationSetNewRelfilenode(wstate.index, RecentXmin);
+
+	if (merge || (btspool->isunique && self->max_dup_errors > 0))
 	{
-		bool	merge = BTReaderInit(&reader, wstate.index);
-
-		elog(DEBUG1, "pg_bulkload: build \"%s\" %s merge (%s wal)",
-			RelationGetRelationName(wstate.index),
-			merge ? "with" : "without",
-			wstate.btws_use_wal ? "with" : "without");
-
-		if (merge)
-		{
-			/* Assign a new file node and merge two streams into it. */
-			setNewRelfilenode(wstate.index, RecentXmin);
-			BULKLOAD_PROFILE_PUSH();
-			_bt_mergeload(&wstate, btspool, &reader, heapRel);
-			BULKLOAD_PROFILE_POP();
-		}
-		else
-		{
-			/* Fast path for newly created index. */
-			_bt_load(&wstate, btspool, NULL);
-		}
-
-		BTReaderTerm(&reader);
+		/* Merge two streams into the new file node that we assigned. */
+		BULKLOAD_PROFILE_PUSH();
+		_bt_mergeload(self, &wstate, btspool, &reader, heapRel);
+		BULKLOAD_PROFILE_POP();
+		BULKLOAD_PROFILE(&prof_merge);
 	}
-	PG_CATCH();
+	else
 	{
-		BTReaderTerm(&reader);
-		PG_RE_THROW();
+		/* Fast path for newly created index. */
+		_bt_load(&wstate, btspool, NULL);
+		BULKLOAD_PROFILE(&prof_index);
 	}
-	PG_END_TRY();
 
-	BULKLOAD_PROFILE(&prof_index_merge_build);
+	BTReaderTerm(&reader);
 }
 
 /*
  * _bt_mergeload - Merge two streams of index tuples into new index files.
  */
 static void
-_bt_mergeload(BTWriteState *wstate, BTSpool *btspool, BTReader *btspool2,
-			  Relation heapRel)
+_bt_mergeload(Spooler *self, BTWriteState *wstate, BTSpool *btspool, BTReader *btspool2, Relation heapRel)
 {
 	BTPageState	   *state = NULL;
 	IndexTuple		itup,
 					itup2;
-	bool			should_free;
+	bool			should_free = false;
 	TupleDesc		tupdes = RelationGetDescr(wstate->index);
 	int				keysz = RelationGetNumberOfAttributes(wstate->index);
 	ScanKey			indexScanKey;
+	ON_DUPLICATE	on_duplicate = self->on_duplicate;
 
 	Assert(btspool != NULL);
-	Assert(btspool2 != NULL);
 
 	/* the preparation of merge */
-	itup = tuplesort_getindextuple(btspool->sortstate, true, &should_free);
+	itup = BTSpoolGetNextItem(btspool, NULL, &should_free);
 	itup2 = BTReaderGetNextItem(btspool2);
 	indexScanKey = _bt_mkscankey_nodata(wstate->index);
 
-	BULKLOAD_PROFILE(&prof_index_merge_build_init);
-
 	for (;;)
 	{
-		bool	check_unique = false;
 		bool	load1 = true;		/* load BTSpool next ? */
+		bool	hasnull;
+		int32	compare;
+
+		if (self->dup_old + self->dup_new > self->max_dup_errors)
+			ereport(ERROR,
+					(errcode(ERRCODE_INTERNAL_ERROR),
+					 errmsg("Maximum duplicate error count exceeded")));
 
 		if (itup2 == NULL)
 		{
@@ -347,77 +433,56 @@ _bt_mergeload(BTWriteState *wstate, BTSpool *btspool, BTReader *btspool2,
 		}
 		else if (itup != NULL)
 		{
-			int		i;
+			compare = compare_indextuple(itup, itup2, indexScanKey,
+										 keysz, tupdes, &hasnull);
 
-			check_unique = btspool->isunique;
-
-			for (i = 1; i <= keysz; i++)
+			if (compare == 0 && !hasnull && btspool->isunique)
 			{
-				ScanKey		entry;
-				Datum		attrDatum1,
-							attrDatum2;
-				bool		isNull1,
-							isNull2;
-				int32		compare;
+				ItemPointerData t_tid2;
 
-				entry = indexScanKey + i - 1;
-				attrDatum1 = index_getattr(itup, i, tupdes, &isNull1);
-				attrDatum2 = index_getattr(itup2, i, tupdes, &isNull2);
-				if (isNull1)
+				/*
+				 * t_tid is update by heap_is_visible(), because use it for an
+				 * index, t_tid backup
+				 */
+				ItemPointerCopy(&itup2->t_tid, &t_tid2);
+
+				/* The tuple pointed by the old index should not be visible. */
+				if (!heap_is_visible(heapRel, &itup->t_tid))
 				{
-					check_unique = false;
-					if (isNull2)
-						compare = 0;		/* NULL "=" NULL */
-					else if (entry->sk_flags & SK_BT_NULLS_FIRST)
-						compare = -1;		/* NULL "<" NOT_NULL */
-					else
-						compare = 1;		/* NULL ">" NOT_NULL */
+					itup = BTSpoolGetNextItem(btspool, itup, &should_free);
 				}
-				else if (isNull2)
+				else if (!heap_is_visible(heapRel, &itup2->t_tid))
 				{
-					check_unique = false;
-					if (entry->sk_flags & SK_BT_NULLS_FIRST)
-						compare = 1;		/* NOT_NULL ">" NULL */
-					else
-						compare = -1;		/* NOT_NULL "<" NULL */
+					itup2 = BTReaderGetNextItem(btspool2);
 				}
 				else
 				{
-					compare = DatumGetInt32(FunctionCall2(&entry->sk_func,
-														  attrDatum1,
-														  attrDatum2));
+					if (on_duplicate == ON_DUPLICATE_KEEP_NEW)
+					{
+						self->dup_old++;
+						remove_duplicate(self, heapRel, itup2,
+							RelationGetRelationName(wstate->index));
+						itup2 = BTReaderGetNextItem(btspool2);
+					}
+					else
+					{
+						ItemPointerCopy(&t_tid2, &itup2->t_tid);
+						self->dup_new++;
+						remove_duplicate(self, heapRel, itup,
+							RelationGetRelationName(wstate->index));
+						itup = BTSpoolGetNextItem(btspool, itup, &should_free);
+					}
+				}
 
-					if (entry->sk_flags & SK_BT_DESC)
-						compare = -compare;
-				}
-				if (compare > 0)
-				{
-					load1 = false;
-					check_unique = false;
-					break;
-				}
-				else if (compare < 0)
-				{
-					check_unique = false;
-					break;
-				}
+				continue;
 			}
+			else if (compare > 0)
+				load1 = false;
 		}
 		else
 			load1 = false;
 
-		if (check_unique)
-		{
-			Assert(load1);
-
-			/* The tuple pointed by the old index should not be visible. */
-			if (heap_is_visible(heapRel, &itup2->t_tid))
-				report_unique_violation(wstate->index, itup);
-
-			/* Discard itup2 and read next */
-			itup2 = BTReaderGetNextItem(btspool2);
-		}
-		BULKLOAD_PROFILE(&prof_index_merge_build_unique);
+		BULKLOAD_PROFILE(&prof_merge_unique);
 
 		/* When we see first tuple, create first index page */
 		if (state == NULL)
@@ -425,24 +490,78 @@ _bt_mergeload(BTWriteState *wstate, BTSpool *btspool, BTReader *btspool2,
 
 		if (load1)
 		{
+			IndexTuple	next_itup = NULL;
+			bool		next_should_free = false;
+
+			for (;;)
+			{
+				/* get next item */
+				next_itup = BTSpoolGetNextItem(btspool, next_itup,
+											   &next_should_free);
+
+				if (!btspool->isunique || next_itup == NULL)
+					break;
+
+				compare = compare_indextuple(itup, next_itup, indexScanKey,
+											 keysz, tupdes, &hasnull);
+				if (compare < 0 || hasnull)
+					break;
+
+				if (compare > 0)
+				{
+					/* shouldn't happen */
+					elog(ERROR, "faild in tuplesort_performsort");
+				}
+
+				/*
+				 * If tupple is deleted by other unique indexes, not visible
+				 */
+				if (!heap_is_visible(heapRel, &next_itup->t_tid))
+				{
+					continue;
+				}
+
+				if (!heap_is_visible(heapRel, &itup->t_tid))
+				{
+					if (should_free)
+						pfree(itup);
+
+					itup = next_itup;
+					should_free = next_should_free;
+					next_should_free = false;
+					continue;
+				}
+
+				/* not unique between input files */
+				self->dup_new++;
+				remove_duplicate(self, heapRel, next_itup,
+								 RelationGetRelationName(wstate->index));
+
+				if (self->dup_old + self->dup_new > self->max_dup_errors)
+					ereport(ERROR,
+							(errcode(ERRCODE_INTERNAL_ERROR),
+							 errmsg("Maximum duplicate error count exceeded")));
+			}
+
 			_bt_buildadd(wstate, state, itup);
+
 			if (should_free)
 				pfree(itup);
-			itup = tuplesort_getindextuple(btspool->sortstate,
-										   true, &should_free);
+
+			itup = next_itup;
+			should_free = next_should_free;
 		}
 		else
 		{
 			_bt_buildadd(wstate, state, itup2);
 			itup2 = BTReaderGetNextItem(btspool2);
 		}
-		BULKLOAD_PROFILE(&prof_index_merge_build_insert);
+		BULKLOAD_PROFILE(&prof_merge_insert);
 	}
 	_bt_freeskey(indexScanKey);
 
 	/* Close down final pages and write the metapage */
 	_bt_uppershutdown(wstate, state);
-	BULKLOAD_PROFILE(&prof_index_merge_build_term);
 
 	/*
 	 * If the index isn't temp, we must fsync it down to disk before it's safe
@@ -464,7 +583,15 @@ _bt_mergeload(BTWriteState *wstate, BTSpool *btspool, BTReader *btspool2,
 		RelationOpenSmgr(wstate->index);
 		smgrimmedsync(wstate->index->rd_smgr, MAIN_FORKNUM);
 	}
-	BULKLOAD_PROFILE(&prof_index_merge_build_flush);
+	BULKLOAD_PROFILE(&prof_merge_term);
+}
+
+static IndexTuple
+BTSpoolGetNextItem(BTSpool *spool, IndexTuple itup, bool *should_free)
+{
+	if (*should_free)
+		pfree(itup);
+	return tuplesort_getindextuple(spool->sortstate, true, should_free);
 }
 
 /**
@@ -500,10 +627,16 @@ BTReaderInit(BTReader *reader, Relation rel)
 	 * will be closed automatically when we assign a new file node.
 	 *
 	 * XXX: It might be better to open the previous relfilenode with
-	 * smgropen *after* setNewRelfilenode.
+	 * smgropen *after* RelationSetNewRelfilenode.
 	 */
 	memset(&reader->smgr, 0, sizeof(reader->smgr));
+#if PG_VERSION_NUM >= 90100
+	reader->smgr.smgr_rnode.node = rel->rd_node;
+	reader->smgr.smgr_rnode.backend =
+		rel->rd_istemp ? MyBackendId : InvalidBackendId;
+#else
 	reader->smgr.smgr_rnode = rel->rd_node;
+#endif
 	reader->smgr.smgr_which = 0;	/* md.c */
 
 	reader->blkno = InvalidBlockNumber;
@@ -631,7 +764,8 @@ BTReaderGetNextItem(BTReader *reader)
 	/*
 	 * If any leaf page isn't read, the state is treated like as EOF 
 	 */
-	Assert(reader->blkno != InvalidBlockNumber);
+	if (reader->blkno == InvalidBlockNumber)
+		return NULL;
 
 	maxoff = PageGetMaxOffsetNumber(reader->page);
 
@@ -672,13 +806,67 @@ BTReaderGetNextItem(BTReader *reader)
 	}
 }
 
+static int
+compare_indextuple(const IndexTuple itup1, const IndexTuple itup2,
+	ScanKey entry, int keysz, TupleDesc tupdes, bool *hasnull)
+{
+	int		i;
+	int32	compare;
+
+	*hasnull = false;
+	for (i = 1; i <= keysz; i++, entry++)
+	{
+		Datum		attrDatum1,
+					attrDatum2;
+		bool		isNull1,
+					isNull2;
+
+		attrDatum1 = index_getattr(itup1, i, tupdes, &isNull1);
+		attrDatum2 = index_getattr(itup2, i, tupdes, &isNull2);
+		if (isNull1)
+		{
+			*hasnull = true;
+			if (isNull2)
+				compare = 0;		/* NULL "=" NULL */
+			else if (entry->sk_flags & SK_BT_NULLS_FIRST)
+				compare = -1;		/* NULL "<" NOT_NULL */
+			else
+				compare = 1;		/* NULL ">" NOT_NULL */
+		}
+		else if (isNull2)
+		{
+			*hasnull = true;
+			if (entry->sk_flags & SK_BT_NULLS_FIRST)
+				compare = 1;		/* NOT_NULL ">" NULL */
+			else
+				compare = -1;		/* NOT_NULL "<" NULL */
+		}
+		else
+		{
+			compare = DatumGetInt32(FunctionCall2(&entry->sk_func,
+												  attrDatum1,
+												  attrDatum2));
+
+			if (entry->sk_flags & SK_BT_DESC)
+				compare = -compare;
+		}
+		if (compare != 0)
+			return compare;
+	}
+
+	return 0;
+}
+
 /*
  * heap_is_visible
+ *
+ * If heap is found, heap_hot_search() update *htid to reference that tuple's
+ * offset number, and return TRUE.  If no match, return FALSE without modifying
+ * tid.
  */
 static bool
 heap_is_visible(Relation heapRel, ItemPointer htid)
 {
-#if PG_VERSION_NUM >= 80300
 	SnapshotData	SnapshotDirty;
 
 	InitDirtySnapshot(SnapshotDirty);
@@ -688,45 +876,106 @@ heap_is_visible(Relation heapRel, ItemPointer htid)
 	 * because we have exclusive lock on the relation. (XXX: Is it true?)
 	 */
 	return heap_hot_search(htid, heapRel, &SnapshotDirty, NULL);
-#else
-	bool			visible;
-	HeapTupleData	htup;
-	Buffer			hbuffer;
-
-	htup.t_self = *htid;
-	visible = heap_fetch(heapRel, SnapshotDirty, &htup, &hbuffer, true, NULL);
-	ReleaseBuffer(hbuffer);
-
-	return visible;
-#endif
 }
 
-/*
- * borrowing from ri_ReportViolation.
- */
 static void
-report_unique_violation(Relation rel, IndexTuple itup)
+remove_duplicate(Spooler *self, Relation heap, IndexTuple itup, const char *relname)
 {
-#define BUFLENGTH	512
-	char		key_names[BUFLENGTH];
-	char		key_values[BUFLENGTH];
-	char	   *name_ptr = key_names;
-	char	   *val_ptr = key_values;
-	int			i;
+	HeapTupleData	tuple;
+	BlockNumber		blknum;
+	BlockNumber		offnum;
+	Buffer			buffer;
+	Page			page;
+	ItemId			itemid;
 
-	TupleDesc	tupdesc = rel->rd_att;
+	blknum = ItemPointerGetBlockNumber(&itup->t_tid);
+	offnum = ItemPointerGetOffsetNumber(&itup->t_tid);
+	buffer = ReadBuffer(heap, blknum);
 
-	for (i = 0; i < tupdesc->natts; i++)
+	LockBuffer(buffer, BUFFER_LOCK_SHARE);
+	page = BufferGetPage(buffer);
+	itemid = PageGetItemId(page, offnum);
+	tuple.t_data = ItemIdIsNormal(itemid)
+		? (HeapTupleHeader) PageGetItem(page, itemid)
+		: NULL;
+	LockBuffer(buffer, BUFFER_LOCK_UNLOCK);
+
+	if (tuple.t_data != NULL)
 	{
-		Datum		value;
-		bool		isnull;
-		char	   *name,
-				   *val;
+		char		   *str;
+		TupleDesc		tupdesc;
 
-		name = NameStr(tupdesc->attrs[i]->attname);
-		value = index_getattr(itup, i + 1, tupdesc, &isnull);
-		if (isnull)
-			val = "null";
+		simple_heap_delete(heap, &itup->t_tid);
+
+		/* output duplicate bad file. */
+		if (self->dup_fp == NULL)
+			if ((self->dup_fp = AllocateFile(self->dup_badfile, "w")) == NULL)
+				ereport(ERROR,
+						(errcode_for_file_access(),
+						 errmsg("could not open duplicate bad file \"%s\": %m",
+								self->dup_badfile)));
+
+		tupdesc = RelationGetDescr(heap);
+		tuple.t_len = ItemIdGetLength(itemid);
+		tuple.t_self = itup->t_tid;
+
+		str = tuple_to_cstring(RelationGetDescr(heap), &tuple);
+		if (fprintf(self->dup_fp, "%s\n", str) < 0 || fflush(self->dup_fp))
+			ereport(ERROR,
+					(errcode_for_file_access(),
+					 errmsg("could not write parse badfile \"%s\": %m",
+							self->dup_badfile)));
+
+		pfree(str);
+	}
+
+	ReleaseBuffer(buffer);
+
+	LoggerLog(WARNING, "Duplicate error Record " int64_FMT
+		": Rejected - duplicate key value violates unique constraint \"%s\"\n",
+		self->dup_old + self->dup_new, relname);
+}
+
+char *
+tuple_to_cstring(TupleDesc tupdesc, HeapTuple tuple)
+{
+	bool		needComma = false;
+	int			ncolumns;
+	int			i;
+	Datum	   *values;
+	bool	   *nulls;
+	StringInfoData buf;
+
+	ncolumns = tupdesc->natts;
+
+	values = (Datum *) palloc(ncolumns * sizeof(Datum));
+	nulls = (bool *) palloc(ncolumns * sizeof(bool));
+
+	/* Break down the tuple into fields */
+	heap_deform_tuple(tuple, tupdesc, values, nulls);
+
+	/* And build the result string */
+	initStringInfo(&buf);
+
+	for (i = 0; i < ncolumns; i++)
+	{
+		char	   *value;
+		char	   *tmp;
+		bool		nq;
+
+		/* Ignore dropped columns in datatype */
+		if (tupdesc->attrs[i]->attisdropped)
+			continue;
+
+		if (needComma)
+			appendStringInfoChar(&buf, ',');
+		needComma = true;
+
+		if (nulls[i])
+		{
+			/* emit nothing... */
+			continue;
+		}
 		else
 		{
 			Oid			foutoid;
@@ -734,29 +983,41 @@ report_unique_violation(Relation rel, IndexTuple itup)
 
 			getTypeOutputInfo(tupdesc->attrs[i]->atttypid,
 							  &foutoid, &typisvarlena);
-			val = OidOutputFunctionCall(foutoid, value);
+			value = OidOutputFunctionCall(foutoid, values[i]);
 		}
 
-		/*
-		 * Go to "..." if name or value doesn't fit in buffer.  We reserve 5
-		 * bytes to ensure we can add comma, "...", null.
-		 */
-		if (strlen(name) >= (key_names + BUFLENGTH - 5) - name_ptr ||
-			strlen(val) >= (key_values + BUFLENGTH - 5) - val_ptr)
+		/* Detect whether we need double quotes for this value */
+		nq = (value[0] == '\0');	/* force quotes for empty string */
+		for (tmp = value; *tmp; tmp++)
 		{
-			sprintf(name_ptr, "...");
-			sprintf(val_ptr, "...");
-			break;
+			char		ch = *tmp;
+
+			if (ch == '"' || ch == '\\' ||
+				ch == '(' || ch == ')' || ch == ',' ||
+				isspace((unsigned char) ch))
+			{
+				nq = true;
+				break;
+			}
 		}
 
-		name_ptr += sprintf(name_ptr, "%s%s", i > 0 ? "," : "", name);
-		val_ptr += sprintf(val_ptr, "%s%s", i > 0 ? "," : "", val);
+		/* And emit the string */
+		if (nq)
+			appendStringInfoChar(&buf, '"');
+		for (tmp = value; *tmp; tmp++)
+		{
+			char		ch = *tmp;
+
+			if (ch == '"' || ch == '\\')
+				appendStringInfoChar(&buf, ch);
+			appendStringInfoChar(&buf, ch);
+		}
+		if (nq)
+			appendStringInfoChar(&buf, '"');
 	}
 
-	ereport(ERROR,
-			(errcode(ERRCODE_UNIQUE_VIOLATION),
-			 errmsg("duplicate key value violates unique constraint \"%s\"",
-					RelationGetRelationName(rel)),
-		errdetail("Key (%s)=(%s) already exists.",
-				  key_names, key_values)));
+	pfree(values);
+	pfree(nulls);
+
+	return buf.data;
 }
